@@ -31,9 +31,11 @@
 *     [12] Guha, Sudipto, et al.
 *          "Robust random cut forest based anomaly detection on streams."
 *          International conference on machine learning. PMLR, 2016.
-* 
+*     [13] Cortes, David.
+*          "Isolation forests: looking beyond tree depth."
+*          arXiv preprint arXiv:2111.11639 (2021).
 *     BSD 2-Clause License
-*     Copyright (c) 2019-2021, David Cortes
+*     Copyright (c) 2019-2022, David Cortes
 *     All rights reserved.
 *     Redistribution and use in source and binary forms, with or without
 *     modification, are permitted provided that the following conditions are met:
@@ -56,6 +58,15 @@
 #include "isotree.hpp" 
 
 
+/* TODO: the method used here is quite slow. From a high level perspective, what the code
+   in this file does is to pass observations down a tree and add +1 for each pair at each
+   step, but this is not actually necessary, as it's also possible to precompute the
+   separation depths between each pair of terminal nodes and then make only one addition
+   per tree for each pair of observations. Should add a 'precompute_distances' method
+   here as a different object class, and perhaps add also the 'isolation kernel'
+   technique for comparison purposes. */
+
+
 /* Calculate distance or similarity between data points
 * 
 * Parameters
