@@ -141,9 +141,20 @@ void split_hplane_recursive(std::vector<IsoHPlane>   &hplanes,
         )
     {
         workspace.col_criterion = ByRange;
-        calc_ranges_all_cols(input_data, workspace, model_params, workspace.node_col_weights.data(),
-                             NULL,
-                             NULL);
+        if (curr_depth == 0 &&
+            (model_params.scoring_metric == BoxedDensity ||
+             model_params.scoring_metric == BoxedDensity2 ||
+             model_params.scoring_metric == BoxedRatio)
+        ) {
+            for (size_t col = 0; col < input_data.ncols_numeric; col++)
+                workspace.node_col_weights[col] = workspace.density_calculator.box_high[col]
+                                                   - workspace.density_calculator.box_low[col];
+        }
+        else {
+            calc_ranges_all_cols(input_data, workspace, model_params, workspace.node_col_weights.data(),
+                                 NULL,
+                                 NULL);
+        }
         workspace.has_saved_stats = false;
     }
 
@@ -349,7 +360,12 @@ void split_hplane_recursive(std::vector<IsoHPlane>   &hplanes,
             if (model_params.missing_action != Fail)
                 hplanes.back().fill_val.assign(workspace.ext_fill_val.begin(), workspace.ext_fill_val.begin() + workspace.ntaken);
 
-            if (model_params.scoring_metric != Depth)
+            if (
+                model_params.scoring_metric != Depth &&
+                model_params.scoring_metric != BoxedDensity &&
+                model_params.scoring_metric != BoxedDensity2 &&
+                model_params.scoring_metric != BoxedRatio
+            )
             {
                 workspace.density_calculator.save_range(workspace.xmin, workspace.xmax);
             }
@@ -504,6 +520,13 @@ void split_hplane_recursive(std::vector<IsoHPlane>   &hplanes,
             workspace.density_calculator.push_density(workspace.xmin, workspace.xmax, hplanes.back().split_point);
         }
 
+        else if (model_params.scoring_metric == BoxedDensity ||
+                 model_params.scoring_metric == BoxedDensity2 ||
+                 model_params.scoring_metric == BoxedRatio)
+        {
+            workspace.density_calculator.push_bdens_ext(hplanes.back(), model_params);
+        }
+
         else
         {
             double pct_tree_left;
@@ -568,7 +591,13 @@ void split_hplane_recursive(std::vector<IsoHPlane>   &hplanes,
     hplanes.emplace_back();
 
     if (impute_nodes != NULL) impute_nodes->emplace_back(hplane_from);
-    if (model_params.scoring_metric != Depth) workspace.density_calculator.pop();
+    if (model_params.scoring_metric == BoxedDensity ||
+        model_params.scoring_metric == BoxedDensity2 ||
+        model_params.scoring_metric == BoxedRatio
+    )
+        workspace.density_calculator.pop_bdens_ext();
+    else if (model_params.scoring_metric != Depth)
+        workspace.density_calculator.pop();
     workspace.st = workspace.split_ix;
     split_hplane_recursive(hplanes,
                            workspace,
@@ -576,6 +605,11 @@ void split_hplane_recursive(std::vector<IsoHPlane>   &hplanes,
                            model_params,
                            impute_nodes,
                            curr_depth + 1);
+    if (model_params.scoring_metric == BoxedDensity ||
+        model_params.scoring_metric == BoxedDensity2 ||
+        model_params.scoring_metric == BoxedRatio
+    )
+        workspace.density_calculator.pop_bdens_ext_right();
 
     return;
 
@@ -625,6 +659,31 @@ void split_hplane_recursive(std::vector<IsoHPlane>   &hplanes,
                 hplanes.back().score = workspace.density_calculator.calc_adj_density();
                 break;
             }
+
+            case BoxedRatio:
+            {
+                hplanes.back().score = workspace.density_calculator.calc_bratio_ext();
+                break;
+            }
+
+            case BoxedDensity:
+            {
+                if (!has_weights)
+                    hplanes.back().score = workspace.density_calculator.calc_bdens_ext(workspace.end - workspace.st + 1, model_params.sample_size);
+                else
+                    hplanes.back().score = workspace.density_calculator.calc_bdens_ext(sum_weight, model_params.sample_size);
+                break;
+            }
+
+            case BoxedDensity2:
+            {
+                if (!has_weights)
+                    hplanes.back().score = workspace.density_calculator.calc_bdens_ext(workspace.end - workspace.st + 1, model_params.sample_size);
+                else
+                    hplanes.back().score = workspace.density_calculator.calc_bdens_ext(sum_weight, model_params.sample_size);
+                break;
+            }
+
         }
 
         /* don't leave any vector initialized */
